import unittest

import numpy as np

from goldilocks.goldilocks import Goldilocks
from goldilocks.strategies import NCounterStrategy

################################################################################
# NOTE Tests following do not test the correctness of regions located by
#      Goldilocks but are instead designed to ensure the program correctly
#      handles inputs and outputs as expected.
################################################################################

DATA = {
    "GroupOne": {
        1: "AAACCCNNNTTTAAACCCGGGNNNAAACCCGGGTTTNNNCCCGGGTTT",
        2: "GCGTNANNGANGGCTANTCTANAGCNNTTTCTNTNNGCANCANTTGNN",
    }
}

class TestGoldilocks(unittest.TestCase):

    def test_invalid_stride(self):
        self.assertRaises(Exception, Goldilocks, NCounterStrategy(), DATA, stride=0)
        self.assertRaises(Exception, Goldilocks, NCounterStrategy(), DATA, stride=-1)
        self.assertRaises(Exception, Goldilocks, NCounterStrategy(), DATA, stride=-1000)

    def test_invalid_length(self):
        self.assertRaises(Exception, Goldilocks, NCounterStrategy(), DATA, length=0)
        self.assertRaises(Exception, Goldilocks, NCounterStrategy(), DATA, length=-1)
        self.assertRaises(Exception, Goldilocks, NCounterStrategy(), DATA, length=-1000)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""A set of classes to enable the handling of aligned block data from psl and 
pslx files."""

__author__ = "Sam Nicholls <sn8@sanger.ac.uk>"
__copyright__ = "Copyright (c) 2012 Genome Research Ltd."
__version__ = 0.50
__license__ = "GNU Lesser General Public License V3"
__maintainer__ = "Sam Nicholls <sam@samnicholls.net>"

#This file is part of Mountain.
#
#Mountain is free software: you can redistribute it and/or modify it under 
#the terms of the GNU Lesser General Public License as published by the Free 
#Software Foundation; either version 3 of the License, or (at your option) any  
#later version.

#This program is distributed in the hope that it will be useful, but WITHOUT 
#ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
#FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more 
#details.

#You should have received a copy of the GNU Lesser General Public License along 
#with this program.  If not, see <http://www.gnu.org/licenses/>.

class PSL(object):
    """Wraps a file handler and provides iterable access over each 'BlockSet' 
    (list of successfully aligned Blocks) in the file"""

    def __init__(self, filepath):
        """Constructs the read only file handler and parses the header"""
        self.handler = open(filepath, 'r')
        self.handler.seek(0)
        line = self.handler.readline().strip()
        if line == "psLayout version 3":
            while(line[0] != '-'):
                line = self.handler.readline()
        else:
            print "Unknown PSL format! Expecting 'psLayout version 3'"
            exit(0)

    def close(self):
        """Close the file handler"""
        self.handler.close()

    def __iter__(self):
        self.handler.seek(0) # Reset the file pointer and re-process header
        return self

    def next(self):
        fields = self.handler.readline().strip().split("\t")
        if len(fields) >= 21:
            if len(fields) > 21:
                block_q_seqs_i = fields[21].upper().split(',')[:-1]
                block_t_seqs_i = fields[22].upper().split(',')[:-1]
            else:
                block_q_seqs_i = [''] * int(fields[17])
                block_t_seqs_i = [''] * int(fields[17])

            return BlockSet(
                match = int(fields[0]),
                mismatch = int(fields[1]),
                repmatch = int(fields[2]),
                n = int(fields[3]),
                q_gap_count = int(fields[4]),
                q_gap_bases = int(fields[5]),
                t_gap_count = int(fields[6]),
                t_gap_bases = int(fields[7]),
                strand = fields[8],
                q_name = fields[9],
                q_size = int(fields[10]),
                q_start = int(fields[11]),
                q_end = int(fields[12]),
                t_name = fields[13],
                t_size = int(fields[14]),
                t_start = int(fields[15]),
                t_end = int(fields[16]),
                block_count = int(fields[17]),
                block_sizes = map(int, fields[18].split(',')[:-1]),
                block_q_starts = map(int, fields[19].split(',')[:-1]),
                block_t_starts = map(int, fields[20].split(',')[:-1]),
                block_q_seqs = block_q_seqs_i,
                block_t_seqs = block_t_seqs_i)
        else:
            raise StopIteration
        

class BlockSet(object):
    """Holds the set of query subsequences and their mappings to the target
    reference sequence for a particular sample. Provides iterable access over 
    each Block in this BlockSet"""

    def __init__(self, **kwds):
        """Construct the BlockSet"""
        #TODO Should provide a way to check all required variables were
        # assigned to ensure future readers create BlockSets of the same form.
        self.__dict__.update(kwds)

    def __iter__(self):
        self.__iter_pointer = -1
        return self

    def next(self):
        self.__iter_pointer += 1
        if self.__iter_pointer < self.block_count:
            return self.get_block(self.__iter_pointer)
        else:
            raise StopIteration

    def get_block(self, i):
        """Get the ith Block in this BlockSet, if it exists"""
        if i < self.block_count:
            return Block(
                i,
                self.block_sizes[i],
                self.block_q_starts[i],
                self.block_t_starts[i],
                self.block_q_seqs[i],
                self.block_t_seqs[i])
        else:
            return None

    def get_all_blocks(self):
        """Get all Blocks in this BlockSet"""
        blocks = []
        for i in range(0, self.block_count):
            blocks.append(self.get_block(i))
        return blocks

class Block(object):
    """An individual alignment; representing a subsequence of the query
    sequence and how it maps to the target reference sequence"""

    def __init__(self, i, size, q_st, t_st, q_seq, t_seq):
        """Construct the Block with its position in the blockset, size, 
        1-index query sequence start, 1-index target sequence start and if the
        read came from a pslx, the aligned query and target subsequences"""
        self.no = i
        self.size = size
        self.q_start = q_st
        self.t_start = t_st
        self.q_seq = q_seq
        self.t_seq = t_seq

    def is_first(self):
        """Check if this Block is the first in the BlockSet"""
        return self.no == 0

    def is_last(self, blockset_len):
        """Check if this Block is the last in the Blockset given the length 
        of the BlockSet"""
        return self.no == (blockset_len - 1)




import math

sequence = "AAACCCNNNTTTAAACCCGGGNNNAAACCCGGGTTTNNNCCCGGGTTT"
sequence = "GCGTNANNGANGGCTANTCTANAGCNNTTTCTNTNNGCANCANTTGNN"
lenseq = len(sequence)

def mark(seq):
    markseq = ["="] * len(seq)
    for i, base in enumerate(seq):
        if base=="N":
            markseq[i] = "*"
    return "".join(markseq)

def entry_contains(subseq):
    count = 0
    for base in subseq:
        if base=="N":
            count += 1
    return count

print
print "#       1 |%s| %d" % (mark(sequence), lenseq)
print "#          %s" % sequence

#TODO Length must be at least 3
stride = 1
length = 3

counts = []
entries = []
entry_size = 0
for i, region_s in enumerate(range(0, lenseq-length+1, stride)):
    region_e = region_s + length - 1

    count = entry_contains(sequence[region_s:region_e+1])
    counts.append(count)
    if length % 2 == 0:
        bucket_str = ("|%s%d%s|" % (" "*int(math.floor((length-2)/2)), count, " "*int(math.floor((length-2)/2)-1)))
    else:
        bucket_str = ("|%s%d%s|" % (" "*int(math.floor((length-2)/2)), count, " "*int(math.floor((length-2)/2))))
    entries.append(bucket_str)
entry_size = len(bucket_str)+stride

# Calculate how many entries can fit on each line beneath the sequence
# when taking the indentation in to consideration
allocations = []
allocated = 0
indent = 0
while allocated < len(entries):
    allocatable = ( int(math.floor((lenseq-indent+stride)/entry_size)) )
    allocations.append(allocatable)
    allocated += allocatable
    indent += 1

# Build the lines
lines = []
for i in range(0, len(allocations)):
    line_i = ""
    for j in range(0, allocations[i]):
        stride_gap = ""
        if j > 0:
            stride_gap = " "*(stride)
        try:
            to_append = stride_gap + entries[ ((j*len(allocations)) + i) ]
            line_i += to_append
        except IndexError:
            # No data left to add
            break
    lines.append(line_i)

for i, l in enumerate(lines):
    print "#%s          %s" % (" "*((i*stride)), "".join(l))

import numpy as np
#for i in reversed(np.argsort(counts)):
#    print "%d\t%d" % (i, counts[i])

buckets = {}
print "EXPECTED_REGIONS = {"
for i, count in enumerate(counts):
    if count not in buckets:
        buckets[count] = []
    buckets[count].append(i)
    print "    %d: %d," % (i, count)
print "}"

print "EXPECTED_BUCKETS = {"
for b in buckets:
    print "    %d: %s," % (b, buckets[b])
print "}"



import unittest

import numpy as np

from goldilocks.goldilocks import Goldilocks
from goldilocks.strategies import NCounterStrategy

################################################################################
# NOTE Tests following are hard coded regression tests whose answers were      #
#      generated and have been checked 'by eye'. Such tests are designed to    #
#      ensure that regressions have not been introduced in to Goldilocks,      #
#      rather than to verify accuracy of its results.                          #
################################################################################

DATA = {
    "GroupOne": {
        1: "AAACCCNNNTTTAAACCCGGGNNNAAACCCGGGTTTNNNCCCGGGTTT",
        2: "GCGTNANNGANGGCTANTCTANAGCNNTTTCTNTNNGCANCANTTGNN",
    },
    "GroupTwo": {
        1: "AAACCCNNNTTTAAACCCGGGNNNAAACCCGGGTTTNNNCCCGGGTTT",
        2: "GCGTNANNGANGGCTANTCTANAGCNNTTTCTNTNNGCANCANTTGNN",
    }
}
STRIDE = 1
LENGTH = 3

# TODO Test group_counts
# TODO Hash Graphs and FASTA
# TODO Test sorting and filtering beyond min
# TODO Test regions for GroupTwo/total
class TestGoldilocksRegression_NCounter(unittest.TestCase):

#  GroupOne
#  CHR 1
#
#       1 |======***============***============***=========| 48
#          AAACCCNNNTTTAAACCCGGGNNNAAACCCGGGTTTNNNCCCGGGTTT
#          |0| |1| |1| |0| |0| |2| |0| |0| |0| |3| |0| |0|
#           |0| |2| |0| |0| |0| |3| |0| |0| |0| |2| |0| |0|
#            |0| |3| |0| |0| |0| |2| |0| |0| |1| |1| |0|
#             |0| |2| |0| |0| |1| |1| |0| |0| |2| |0| |0|
#

#  GroupOne
#  CHR 2
#
#       1 |====*=**==*=====*====*===**=====*=**===*==*===**| 48
#          GCGTNANNGANGGCTANTCTANAGCNNTTTCTNTNNGCANCANTTGNN
#          |0| |2| |1| |0| |1| |1| |2| |0| |2| |0| |1| |1|
#           |0| |2| |1| |0| |0| |1| |2| |0| |2| |1| |1| |2|
#            |1| |2| |1| |1| |0| |0| |1| |1| |2| |1| |1|
#             |1| |1| |0| |1| |1| |1| |0| |1| |1| |1| |0|
#
    @classmethod
    def setUpClass(cls):
        cls.g = Goldilocks(NCounterStrategy(), DATA, stride=STRIDE, length=LENGTH)
        cls.EXPECTED_REGIONS = {
            1: {
                0: 0,
                1: 0,
                2: 0,
                3: 0,
                4: 1,
                5: 2,
                6: 3,
                7: 2,
                8: 1,
                9: 0,
                10: 0,
                11: 0,
                12: 0,
                13: 0,
                14: 0,
                15: 0,
                16: 0,
                17: 0,
                18: 0,
                19: 1,
                20: 2,
                21: 3,
                22: 2,
                23: 1,
                24: 0,
                25: 0,
                26: 0,
                27: 0,
                28: 0,
                29: 0,
                30: 0,
                31: 0,
                32: 0,
                33: 0,
                34: 1,
                35: 2,
                36: 3,
                37: 2,
                38: 1,
                39: 0,
                40: 0,
                41: 0,
                42: 0,
                43: 0,
                44: 0,
                45: 0,
            },
            2: {
                0: 0,
                1: 0,
                2: 1,
                3: 1,
                4: 2,
                5: 2,
                6: 2,
                7: 1,
                8: 1,
                9: 1,
                10: 1,
                11: 0,
                12: 0,
                13: 0,
                14: 1,
                15: 1,
                16: 1,
                17: 0,
                18: 0,
                19: 1,
                20: 1,
                21: 1,
                22: 0,
                23: 1,
                24: 2,
                25: 2,
                26: 1,
                27: 0,
                28: 0,
                29: 0,
                30: 1,
                31: 1,
                32: 2,
                33: 2,
                34: 2,
                35: 1,
                36: 0,
                37: 1,
                38: 1,
                39: 1,
                40: 1,
                41: 1,
                42: 1,
                43: 0,
                44: 1,
                45: 2,
            }
        }

        cls.EXPECTED_REGION_TOTALS = []
        for j in cls.EXPECTED_REGIONS:
            for i in cls.EXPECTED_REGIONS[j]:
                cls.EXPECTED_REGION_TOTALS.append(cls.EXPECTED_REGIONS[j][i]*2)

        cls.EXPECTED_BUCKETS = {
            0: [0, 1, 2, 3, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 24, 25, 26,
                    27, 28, 29, 30, 31, 32, 33, 39, 40, 41, 42, 43, 44, 45,
                        46, 47, 57, 58, 59, 63, 64, 68, 73, 74, 75, 82, 89],
            1: [4, 8, 19, 23, 34, 38,
                        48, 49, 53, 54, 55, 56, 60, 61, 62, 65, 66, 67, 69,
                        72, 76, 77, 81, 83, 84, 85, 86, 87, 88, 90],
            2: [5, 7, 20, 22, 35, 37,
                        50, 51, 52, 70, 71, 78, 79, 80, 91],
            3: [6, 21, 36],
        }
        cls.EXPECTED_REGION_COUNT = len(cls.EXPECTED_REGIONS[1]) + len(cls.EXPECTED_REGIONS[2])

    def test_number_regions(self):
        self.assertEqual(self.EXPECTED_REGION_COUNT, len(self.g.regions.items()))

    # TODO Technically these check the metadata is correct and not the *actual* regions
    def test_region_lengths(self):
        number_comparisons = 0
        # Ensure ALL meet LENGTH
        for region_i, region_data in self.g.regions.items():
            number_comparisons += 1
            self.assertEqual(LENGTH, len(range(region_data["pos_start"], region_data["pos_end"])) + 1)
        self.assertEqual(self.EXPECTED_REGION_COUNT, number_comparisons)

    # TODO Technically these check the metadata is correct and not the *actual* regions
    def test_region_stride(self):
        number_comparisons = 0
        # Ensure regions begin at right STRIDE
        for region_i, region_data in self.g.regions.items():
            number_comparisons += 1
            expected_start = 1 + (region_data["ichr"] * STRIDE)
            self.assertEqual(expected_start, region_data["pos_start"])

            # -1 as the region includes the start element
            expected_end = (expected_start - 1) + LENGTH
            self.assertEqual(expected_end, region_data["pos_end"])
        self.assertEqual(self.EXPECTED_REGION_COUNT, number_comparisons)

###############################################################################
# \/ NOTE Tests below depend only on GroupOne                              \/ #
###############################################################################

    def test_content_regions(self):
        number_comparisons = 0
        for region_i, region_data in self.g.regions.items():
            number_comparisons += 1
            self.assertEqual(self.EXPECTED_REGIONS[region_data["chr"]][region_data["ichr"]], region_data["group_counts"]["GroupOne"]["default"])
        self.assertEqual(self.EXPECTED_REGION_COUNT, number_comparisons)

    def test_number_buckets(self):
        self.assertEqual(len(self.EXPECTED_BUCKETS), len(self.g.group_buckets["GroupOne"]))
        total_in_bucket = 0
        total_in_bucket_g = 0

        for b in self.EXPECTED_BUCKETS:
            total_in_bucket += len(self.EXPECTED_BUCKETS[b])
        for b in self.g.group_buckets["GroupOne"]:
            total_in_bucket_g += len(self.g.group_buckets["GroupOne"][b])

        self.assertEqual(total_in_bucket, total_in_bucket_g)

    def test_content_buckets(self):
        number_comparisons = 0
        for bucket, content in self.g.group_buckets["GroupOne"].items():
            number_comparisons += 1
            self.assertEqual(sorted(self.EXPECTED_BUCKETS[bucket]), sorted(content))
        self.assertEqual(len(self.EXPECTED_BUCKETS), number_comparisons)

###############################################################################
# /\ NOTE Tests above depend only on GroupOne                              /\ #
###############################################################################

    def test_sort_min(self):
        number_comparisons = 0
        regions_min = self.g._filter("min")

        # Merge sort required for stability
        EXPECTED_MIN_REGION_ORDER = np.argsort(self.EXPECTED_REGION_TOTALS, kind="mergesort")

        for i, region in enumerate(regions_min):
            self.assertEquals(EXPECTED_MIN_REGION_ORDER[i], region['id'])
            number_comparisons += 1
        self.assertEqual(self.EXPECTED_REGION_COUNT, number_comparisons)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""A set of classes to enable the handling of FASTA files and libraries"""

__author__ = "Sam Nicholls <sn8@sanger.ac.uk>"
__copyright__ = "Copyright (c) 2012 Genome Research Ltd."
__version__ = 0.15
__license__ = "GNU Lesser General Public License V3"
__maintainer__ = "Sam Nicholls <sam@samnicholls.net>"

#This file is part of Mountain.
#
#Mountain is free software: you can redistribute it and/or modify it under 
#the terms of the GNU Lesser General Public License as published by the Free 
#Software Foundation; either version 3 of the License, or (at your option) any  
#later version.

#This program is distributed in the hope that it will be useful, but WITHOUT 
#ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
#FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more 
#details.

#You should have received a copy of the GNU Lesser General Public License along 
#with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

CONVERT_LINES = re.compile("\r?\n|\r")

class FASTA_Index(object):
    def __init__(self, filepath):
        index_fp = open(filepath, 'r')

        self.book_index = {}
        self.book_lengths = {}

        for line in index_fp:
            fields = line.strip().split("\t")

            curr_fasta = {
                "name": fields[0],
                "length": fields[1],
                "file_pos": fields[2],
                "line_bases": fields[3],
                "line_bytes": fields[4]
            }

            self.book_index[fields[0]] = curr_fasta
            self.book_lengths[fields[0]] = fields[1]

        index_fp.close()

class FASTA_Library(object):
    """Wraps a file handler and provides iterable access over each FASTA"""

    def __init__(self, filepath, index=None):
        """Constructs the read only file handler and parses the header"""
        self.max_chr_sizes = {}

        self.handler = open(filepath, 'r')
        self.books = {}
        self.process_file()

        if index:
            self.library = FASTA_Index(index)
        else:
            self.library = None
            self.build_library()

    def build_library(self):
        self.handler.seek(0)
        #NOTE Cannot rely on file iterator as this creates a read-ahead buffer
        #     to improve efficiency which alters the location of the file pointer.
        for line in iter(self.handler.readline, ''):
            if line[0] == '>':
                #TODO Hash readname or something?
                current_book = line.strip()[1:]
                self.books[current_book] = {}
                self.books[current_book]['pos'] = self.handler.tell()
                self.books[current_book]['len'] = 0
                self.max_chr_sizes[current_book] = 0
                print "[READ] %s@%d" % (line.strip()[1:], self.handler.tell())
            else:
                self.books[current_book]['len'] += len(line.strip())
                self.max_chr_sizes[current_book] += len(line.strip())
        self.handler.seek(0)

    def seek_fasta(self, book):
        if book not in self.books:
            raise Exception("Unknown Book: %s" % book)
        self.handler.seek(self.books[book]['pos'], 0)

    def fetch_n_bases(self, num_bases):
        #TODO Catch start of next book
        seqbuff = ""
        while len(seqbuff) < num_bases:

            curr_line = self.handler.readline()
            while curr_line == "\n":
#???
                curr_line = self.handler.readline()
            curr_line_s = curr_line.strip()

            if (len(seqbuff) + len(curr_line_s)) <= num_bases:
                next_base = curr_line_s
            else:
                underflow = num_bases - len(seqbuff)
                overflow = (len(seqbuff)+len(curr_line_s)) - num_bases

                # Incl /n
                if curr_line[-1] == "\n":
                    overflow += 1
                next_base = curr_line[:underflow]
                self.handler.seek(-overflow, 1)

            if len(next_base) == 0:
                print "[WARN] EOF reached fetching %d bases. %d returned." % (num_bases, len(seqbuff))
                return seqbuff

            seqbuff += next_base
        return seqbuff

    def process_file(self):
        """Ensure the file contains FASTA sequences and remove any headers"""
        self.handler.seek(0)
        line = self.handler.readline().strip()

        # Remove any possible headers
        while(line[0] != '>'):
            line = self.handler.readline()

            if not line:
                print ("Reached EOF without finding a readname. "
                    "No FASTA sequences found.")
                exit(0)

        self.current_readname = line
        self.has_next = 1

    def close(self):
        """Close the file handler"""
        self.handler.close()

    def __iter__(self):
        self.process_file() #Reset file pointer and re-process file header
        return self

    def next(self):
        if self.has_next:
            name = self.current_readname.strip()
            self.has_next = 0
        else:
            raise StopIteration

        seq = ""
        line = self.handler.readline()
        while line and self.has_next == 0:
            if line[0] != '>':
                seq += CONVERT_LINES.sub("", line)
                line = self.handler.readline()
            else:
                self.current_readname = line
                self.has_next = 1

        return FASTA(name, seq)

    def get_next(self):
        """Return the next FASTA sequence in the library, if there is one.
        Typically used in place of an iterator where you know there is only
        one sample inside the file"""
        if self.has_next:
            return self.next()
        else:
            return None

class FASTA(object):
    """Simple encapsulation of a FASTA sequence"""

    def __init__(self, name, seq):
        """Construct the FASTA with a readname and sequence"""
        self.readname = name
        self.seq = seq

    def __iter__(self):
        """Provide an iterator over the bases in this FASTA sequence"""
        return iter(self.seq)

    def __len__(self):
        """Return the length of this FASTA seqeunce"""
        return len(self.seq)




#!/usr/bin/env python
"""Standard Bases and IUPAC Lookup Dictionary"""

__author__ = "Sam Nicholls <sn8@sanger.ac.uk>"
__copyright__ = "Copyright (c) 2012 Genome Research Ltd."
__version__ = 1.0
__license__ = "GNU Lesser General Public License V3"
__maintainer__ = "Sam Nicholls <sam@samnicholls.net>"

#This file is part of Mountain.
#
#Mountain is free software: you can redistribute it and/or modify it under
#the terms of the GNU Lesser General Public License as published by the Free
#Software Foundation; either version 3 of the License, or (at your option) any
#later version.
#
#This program is distributed in the hope that it will be useful, but WITHOUT
#ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
#details.
#
#You should have received a copy of the GNU Lesser General Public License along
#with this program.  If not, see <http://www.gnu.org/licenses/>.

# Standard Bases
STANDARD_BASES = ['G', 'C', 'T', 'A']

# Lookup dict for reconciling IUPAC codes to standard bases.
# Constructed with data from:
# http://en.wikipedia.org/wiki/Nucleic_acid_notation#IUPAC_notation
IUPAC = {
    'R': ['A', 'G'],
    'Y': ['C', 'T'],
    'S': ['G', 'C'],
    'W': ['A', 'T'],
    'K': ['G', 'T'],
    'M': ['A', 'C'],
    'B': ['C', 'G', 'T'],
    'D': ['A', 'G', 'T'],
    'H': ['A', 'C', 'T'],
    'V': ['A', 'C', 'G'],
    'N': ['A', 'C', 'T', 'G']}



__author__ = "Sam Nicholls <sn8@sanger.ac.uk>"
__copyright__ = "Copyright (c) Sam Nicholls"
__version__ = "0.0.31"
__maintainer__ = "Sam Nicholls <sam@samnicholls.net>"

import numpy as np
from math import floor, ceil

class CandidateList(list):
    """A list defining its own tab-delimited table string representation when
    printed by a user. Provides other utility methods for generating other useful
    outputs including exporting sequences to FASTA."""
    def __init__(self, g, *args):
        self.__goldilocks = g
        list.__init__(self, *args)

    def __repr__(self):
        str_rep = "ID\tVAL\tCHR\tPOSITIONS (INC.)\n"
        for region in self:
            str_rep += ("%d\t%s\t%s\t%10d - %10d\n" % (region["id"],
                                            region["group_counts"]["total"],
                                            region["chr"],
                                            region["pos_start"],
                                            region["pos_end"],
            ))
        return str_rep

    #TODO Export to file not stdout!
    def export_fasta(self, group):
        """Export all regions held in the CandidateList in FASTA format."""
        for region in self:
            print(">%s|Chr%s|Pos%d:%d" % (group, region["chr"], region["pos_start"], region["pos_end"]))
            print(self.__goldilocks.groups[group][region["chr"]][region["pos_start"]:region["pos_end"]+1])


# TODO Generate database of regions with stats... SQL/SQLite
#      - Probably more of a wrapper script than core-functionality: goldib
class Goldilocks(object):
    """Facade class responsible for conducting a census of provided genomic regions
    using the given strategy and provides an interface via _filter to query results
    for given criteria and return a CandidateList."""

    def __load_chromosome(self, arr, data, track):
        return self.strategy.prepare(arr, data, track)

    def __init__(self, strategy, data, is_seq=True, length=None, stride=1, med_window=12.5):

        self.strategy = strategy# Search strategy

        self.chr_max_len = {}   # Map chromosomes to the largest variant position
                                # seen across all files

        self.groups = {}        # For each group stores a dict with chromosome
                                # numbers as keys with lists of variant positions
                                # as values

        self.group_buckets = {} # For each group holds keys of region sizes with
                                # values a list of region_i of that size

        self.group_counts = {}  # Holds a list of each region size seen for each
                                # group for calculating quantiles etc.

        self.regions = {}       # Stores information on each region checked

        self.MED_WINDOW = med_window # Middle 25%, can also be overriden later

        # Read data
        self.groups = data
        self.max_chr_max_len = None

        for group in self.groups:
            #TODO Catch duplicates etc..
            self.group_buckets[group] = {}
            self.group_counts[group] = {}

            for chrom in self.groups[group]:
                # Remember to exclude 0-index
                if is_seq:
                    # TODO This is awful.
                    # Prepend _ to sequence to force 1-index
                    if self.groups[group][chrom][0] != "_":
                        self.groups[group][chrom] = "_" + self.groups[group][chrom]

                    len_current_seq = len(self.groups[group][chrom]) - 1
                else:
                    len_current_seq = sorted(self.groups[group][chrom])[-1]

                #TODO Should we not be looking for min length?
                if chrom not in self.chr_max_len:
                    self.chr_max_len[chrom] = len_current_seq

                if len_current_seq > self.chr_max_len[chrom]:
                    self.chr_max_len[chrom] = len_current_seq

                if self.max_chr_max_len is None:
                    self.max_chr_max_len = len_current_seq
                elif len_current_seq > self.max_chr_max_len:
                    self.max_chr_max_len = len_current_seq

        self.group_buckets["total"] = {}
        self.group_counts["total"] = {}

        # Ensure stride and length are valid, is a length has not been provided
        # use the size of the largest chromosome divided by 10.
        if stride < 1:
            raise Exception("[FAIL] Stride must be at least 1 base wide.")
        self.STRIDE = stride

        # TODO Somewhat arbitrary...
        if length is None:
            length = self.max_chr_max_len / 10

        if length < 1:
            raise Exception("[FAIL] Length must be at least 1 base wide.")
        self.LENGTH = length

        # Conduct a census of the regions on each chromosome using the user
        # defined length and stride. Counting the number of variants present for
        # each group.
        regions = {}
        region_i = 0
        for chrno, size in sorted(self.chr_max_len.items()):
            chros = {}
            for group in self.groups:
                chros[group] = {}
                for track in self.strategy.TRACKS:
                    chro = np.zeros(size+1, np.int8)
                    chros[group][track] = self.__load_chromosome(chro, self.groups[group][chrno], track)

            print("[SRCH] Chr:%s" % str(chrno))
            # Ignore 0 position
            for i, region_s in enumerate(range(1, size+1-self.LENGTH+1, self.STRIDE)):
                region_e = region_s + self.LENGTH - 1
                regions[region_i] = {
                    "ichr": i,
                    "group_counts": {"total": {}},
                    "chr": chrno,
                    "pos_start": region_s,
                    "pos_end": region_e
                }

                for group in self.groups:
                    for track in self.strategy.TRACKS:
                        value = self.strategy.evaluate(chros[group][track][region_s:region_e+1], track)

                        if group not in regions[region_i]["group_counts"]:
                            regions[region_i]["group_counts"][group] = {}
                        if track not in regions[region_i]["group_counts"]["total"]:
                            regions[region_i]["group_counts"]["total"][track] = 0

                        regions[region_i]["group_counts"][group][track] = value
                        regions[region_i]["group_counts"]["total"][track] += value

                        # TODO Should we be ignoring these regions if they are empty?
                        if value not in self.group_buckets[group]:
                            # Add this particular number of variants as a bucket
                            self.group_buckets[group][value] = []

                        if value not in self.group_buckets["total"]:
                            self.group_buckets["total"][value] = []

                        # Add the region id to the bucket
                        self.group_buckets[group][value].append(region_i)
                        self.group_buckets["total"][value].append(region_i)

                    # TODO Config option to include 0 in filter metrics
#                   if value > 0:
                        # Append the number of variants counted in this region
                        # for this group to a list
                        if track not in self.group_counts["total"]:
                            self.group_counts["total"][track] = []
                        if track not in self.group_counts[group]:
                            self.group_counts[group][track] = []

                        self.group_counts[group][track].append(value)

                        try:
                            self.group_counts["total"][track][region_i] += value
                        except IndexError:
                            self.group_counts["total"][track].append(value)

                region_i += 1
        self.regions = regions

    def __apply_filter_func(self, func, lower_window, upper_window, group, actual, track):
        valid_funcs = [
                "median",
                "mean",
                "max",
                "min"
        ]

        target = None
        if func.lower() == "median":
            if actual:
                q_low  = np.percentile(np.asarray(self.group_counts[group][track]), 50) - lower_window
                q_high = np.percentile(np.asarray(self.group_counts[group][track]), 50) + upper_window
            else:
                q_low  = np.percentile(np.asarray(self.group_counts[group][track]), 50 - lower_window)
                q_high = np.percentile(np.asarray(self.group_counts[group][track]), 50 + upper_window)
            target = np.percentile(np.asarray(self.group_counts[group][track]), 50)
        if func.lower() == "mean":
            if actual:
                q_low  = np.percentile(np.asarray(self.group_counts[group][track]), 50) - lower_window
                q_high = np.percentile(np.asarray(self.group_counts[group][track]), 50) + upper_window
            else:
                q_low  = np.percentile(np.asarray(self.group_counts[group][track]), 50 - lower_window)
                q_high = np.percentile(np.asarray(self.group_counts[group][track]), 50 + upper_window)
            target = np.percentile(np.asarray(self.group_counts[group][track]), 50)
        elif func.lower() == "max":
            q_high = np.percentile(np.asarray(self.group_counts[group][track]), 100)
            if actual:
                q_low  = np.percentile(np.asarray(self.group_counts[group][track]), 100) - lower_window
            else:
                q_low  = np.percentile(np.asarray(self.group_counts[group][track]), 100 - lower_window)
            target = q_high
        elif func.lower() == "min":
            q_low = np.percentile(np.asarray(self.group_counts[group][track]), 0)
            if actual:
                q_high = np.percentile(np.asarray(self.group_counts[group][track]), 0) + upper_window
            else:
                q_high = np.percentile(np.asarray(self.group_counts[group][track]), 0 + upper_window)
            target = q_low
        else:
            raise NotImplementedException("[FAIL] Function '%s' not supported" % func.lower())
        return q_low, q_high, target


    def __check_exclusions(self, exclusions, region_dict, use_and=False):
        if exclusions is None or len(exclusions) == 0:
            return False

        def exclude_chro(region_dict, chr_bool):
            if chr_bool is True:
                return True
            return False

        def exclude_start(region_dict, operand, position):
            if operand < 0:
                if region_dict["pos_start"] <= position:
                    return True
            elif operand > 0:
                if region_dict["pos_start"] >= position:
                    return True

            return False

        def exclude_end(region_dict, operand, position):
            if operand < 0:
                if region_dict["pos_end"] <= position:
                    return True
            elif operand > 0:
                if region_dict["pos_end"] >= position:
                    return True

            return False

        # Chrom specific exclusions override overall (might be unexpected)
        if region_dict["chr"] in exclusions:
            to_apply = exclusions[region_dict["chr"]]
        elif 0 in exclusions:
            to_apply = exclusions[0]
        else:
            # No exclusions to apply to this region
            return False

        for name in to_apply:
            #TODO Could probably improve with a dict of funcs...
            ret = False
            if name == "chr":
                ret = exclude_chro(region_dict, to_apply["chr"])
            elif name == "start_lte":
                ret = exclude_start(region_dict, -1, to_apply["start_lte"])
            elif name == "start_gte":
                ret = exclude_start(region_dict, 1, to_apply["start_gte"])
            elif name == "end_lte":
                ret = exclude_end(region_dict, -1, to_apply["end_lte"])
            elif name == "end_gte":
                ret = exclude_end(region_dict, 1, to_apply["end_gte"])
            else:
                #TODO Invalid option
                pass

            if use_and:
                # Require all exclusions to be true...
                if ret is False:
                    return False
            else:
                # If we're not waiting on all conditions, we can exclude on the first
                if ret is True:
                    return True

        if use_and:
            # If we didn't bail on a previous false, all conditions must be satisfied
            return True
        return False

    # TODO Pretty hacky at the moment... Just trying some things out!
    def _filter(self, func="median", track="default", actual_distance=None, percentile_distance=None,
            direction=0, group=None, limit=0, exclusions=None, use_and=False):
        distance = None
        actual = False
        if actual_distance is not None and percentile_distance is not None:
            raise Exception("[FAIL] Cannot filter by both actual_distance and percentile_difference. Select one.")

        if actual_distance is not None:
            distance = actual_distance
            actual = True

        if percentile_distance is not None:
            distance = percentile_distance

        if group is None:
            group = "total"

        candidates = []
        if distance is not None:
            upper_window = float(distance) / 2
            lower_window = float(distance) / 2

            if direction > 0:
                upper_window = upper_window * 2
                lower_window = 0.0
            elif direction < 0:
                lower_window = lower_window * 2
                upper_window = 0.0

            q_low, q_high, target = self.__apply_filter_func(func, upper_window, lower_window, group, actual, track)
        else:
            #TODO Pretty ugly.
            q_low, q_high, target = self.__apply_filter_func(func, 0, 0, group, actual, track)
            q_low  = min(np.asarray(self.group_counts[group][track]))
            q_high = max(np.asarray(self.group_counts[group][track]))

        # For each "number of variants" bucket: which map the number of variants
        # seen in a region, to all regions that contained that number of variants
        print("[NOTE] Filtering values between %.2f and %.2f (inclusive)" % (floor(q_low), ceil(q_high)))
        for bucket in self.group_buckets[group]:
            if bucket >= floor(q_low) and bucket <= ceil(q_high):
                # Append all region data structures within the desired range
                # to the list of candidates for enrichment
                candidates += self.group_buckets[group][bucket]

        num_selected = 0
        num_excluded = 0
        num_total = 0

        filtered = CandidateList(self)
        for region in sorted(self.regions,
                        key=lambda x: (abs(self.regions[x]["group_counts"][group][track] - target),
                            self.regions[x]["group_counts"][group][track])):
            num_total += 1
            if region in candidates:
                num_selected += 1
                if not self.__check_exclusions(exclusions, self.regions[region], use_and):
                    self.regions[region]["id"] = region
                    filtered.append(self.regions[region])
                else:
                    num_excluded += 1

        # Return the top N elements if desired
        # TODO Report total, selected, selected-excluded and selected-filtered
        if limit:
            filtered = CandidateList(self, filtered[0:limit])

        print("[NOTE] %d processed, %d match search criteria, %d excluded, %d limit" %
                (num_total, num_selected, num_excluded, limit))

        return filtered

    def plot(self, group=None, track="default", ylim=None, save_to=None, annotation=None):
        """Represent censused regions in a plot using matplotlib."""

        import matplotlib.pyplot as plt

        if group is None:
            group = "total"
            fig = plt.subplot(1,1,1)

            num_regions = len(self.regions)
            num_counts = [self.regions[x]["group_counts"][group][track] for x in sorted(self.regions)]

            max_val = max(num_counts)
            plt.scatter(range(0, num_regions), num_counts, c=num_counts, marker='o')
            plt.plot(range(0, num_regions), num_counts, "black")
            plt.axis([0, num_regions, 0, max_val])
            plt.ylabel(self.strategy.AXIS_TITLE)
            if ylim:
                plt.ylim(ylim)
        else:
            fig, ax = plt.subplots(len(self.groups[group]),1, sharex=True, squeeze=False)

            max_val = 0
            for i, chrom in enumerate(self.groups[group]):

                num_counts = [self.regions[x]["group_counts"][group][track] for x in sorted(self.regions) if self.regions[x]["chr"] == chrom]
                num_regions = len(num_counts)

                if max(num_counts) > max_val:
                    max_val = max(num_counts)

                ax[i,0].plot(range(0, num_regions), num_counts, label="g"+str(chrom))
                ax[i,0].text(
                    1.05, 0.5, ("Chr#"+str(chrom)), transform=ax[i,0].transAxes,
                    rotation=270, fontsize=12, va='top',
                    horizontalalignment='center', verticalalignment='center'
                )

                if ylim:
                    ax[i,0].set_ylim(ylim)

            # Y axis label
            fig.text(
                .05, 0.5, self.strategy.AXIS_TITLE, rotation='vertical',
                horizontalalignment='center', verticalalignment='center'
            )

        plt.xlabel("Region# (150bp)")
        plt.suptitle('%s-%s' % (group, track), fontsize=16)

        if annotation:
            plt.annotate(annotation, xy=(.5, 1.03),  xycoords='axes fraction', ha='center', va='center', fontsize=11)

        if save_to:
            plt.savefig(save_to)
            plt.close()
        else:
            plt.show()

    #TODO Export to file not stdout!
    def export_meta(self, group, sep=","):
        """Export census metadata to stdout with the following header: id, track1,
        [track2 ... trackN], chr, pos_start, pos_end. Accepts a seperator but
        defaults to a comma-delimited table."""

        tracks = sorted(self.strategy.TRACKS)
        tracks_header = sep.join(tracks)

        print(sep.join(["id", tracks_header, "chr", "pos_start", "pos_end"]))
        for r in self.regions:
            region = self.regions[r]
            values_string = ""
            for t in tracks:
                values_string += str(region["group_counts"][group][t])
                values_string += sep
            values_string = values_string[:-1]
            print(sep.join([
                str(r),
                values_string,
                str(region["chr"]),
                str(region["pos_start"]),
                str(region["pos_end"])]
            ))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import setuptools

README = open('README.rst').read()
CHANGELOG = open('CHANGELOG.rst').read()

requirements = [
    "numpy"
]

test_requirements = [
    "tox",
    "pytest",
    "nose",
    "python-coveralls",
]

setuptools.setup(
    name="goldilocks",
    version="0.0.31",
    url="https://github.com/samstudio8/goldilocks",

    description="Locating genomic regions that are \"just right\".",
    long_description=README + '\n\n' + CHANGELOG,

    author="Sam Nicholls",
    author_email="sam@samnicholls.net",

    maintainer="Sam Nicholls",
    maintainer_email="sam@samnicholls.net",

    packages=setuptools.find_packages(),
    include_package_data=True,

    install_requires=requirements,

    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],

    test_suite='tests',
    tests_require=test_requirements
)



import numpy as np

class BaseStrategy(object):
    """ """
    def __init__(self, tracks=None, title=""):
        if tracks is None:
            tracks = ["default"]

        self.TRACKS = tracks
        self.AXIS_TITLE = title

    def prepare(self, arr, data, track):
        """Populate elements in a given iterable 'arr' (typically a numpy
        array) following some processing strategy on genomic sequence 'data'.
        Tracks can be used to further inform the strategy on what
        behaviour to use (for example a nucleotide counting strategy would need
        a 'track' for each base to be counted)."""
        raise NotImplementedError("strategy.prepare")

    def evaluate(self, region, track):
        """Evaluate the contents of a given iterable 'region' (typically a numpy
        array) as prepared by this strategy. The simplest strategies will sum
        the binary flags over the array."""
        raise NotImplementedError("strategy.evaluate")


class KMerCounterStrategy(BaseStrategy):

    def __init__(self, kmers):
        super(KMerCounterStrategy, self).__init__(tracks=kmers, title="Motif Count")

    def prepare(self, arr, data, track):
        import re

        # Populate the region array with 1 for the start position of desired K-Mer
        #TODO Use KMP?
        for location in [m.start() for m in re.finditer(track, data)]:
            arr[location] = 1
        return arr

    def evaluate(self, region, track):
        # Need to allow for cases where the K-mer present flag has been placed
        # at the very end of the region (and thus the region merely contains only
        # the first base of the desired K-mer
        return np.sum(region[:-(len(track)-1)])

class VariantCounterStrategy(BaseStrategy):

    def __init__(self, tracks=None):
        super(VariantCounterStrategy, self).__init__(title="Variant Count")

    def prepare(self, arr, data, track):
        """Return a NumPy array containing 1 for position elements where a variant
        exists and 0 otherwise."""
        # Populate the chromosome array with 1 for each position a variant exists
        for variant_loc in data:
            arr[variant_loc] = 1
        return arr

    def evaluate(self, region, track):
        return np.sum(region)


class GCRatioStrategy(BaseStrategy):

    def __init__(self, tracks=None):
        super(GCRatioStrategy, self).__init__(title="GC Ratio")

    def prepare(self, arr, data, track):
        # Populate the region array with 1 for each position a GC base exists
        #for location, base in enumerate(data):
        #    base_u = base.upper()
        #    if base_u == "G" or base_u == "C":
        #        arr[location] = 1
        #return arr
        import re
        for location in [m.start() for m in re.finditer("[GC]", data)]:
            arr[location] = 1
        return arr

    def evaluate(self, region, track):
        return float(np.sum(region))/len(region)


class NCounterStrategy(BaseStrategy):

    def __init__(self, tracks=None):
        super(NCounterStrategy, self).__init__(title="N Base Count")

    def prepare(self, arr, data, track):
        # Populate the region array with 1 for each position a missing base exists
        for location, base in enumerate(data):
            if base.upper() == "N":
                arr[location] = 1
        return arr

    def evaluate(self, region, track):
        return np.sum(region)

